@@ -3,35 +3,38 @@
  *  Module    : group.c
  *  Author    : I. Lea & R. Skrenta
  *  Created   : 1991-04-01
- *  Updated   : 2017-10-24
+ *  Updated   : 2024-12-02
  *  Notes     :
  *
- * Copyright (c) 1991-2019 Iain Lea <iain@bricbrac.de>, Rich Skrenta <skrenta@pbm.com>
+ * Copyright (c) 1991-2025 Iain Lea <iain@bricbrac.de>, Rich Skrenta <skrenta@pbm.com>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
+ *
+ * 1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions and the following disclaimer.
+ *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote
- *    products derived from this software without specific prior written
- *    permission.
  *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
- * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
- * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
- * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
- * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
- * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * 3. Neither the name of the copyright holder nor the names of its
+ *    contributors may be used to endorse or promote products derived from
+ *    this software without specific prior written permission.
+ *
+ * * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
+ * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
+ * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+ * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
+ * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+ * POSSIBILITY OF SUCH DAMAGE.
  */
 
 
@@ -58,7 +61,7 @@ static struct t_fmt grp_fmt;
 static int do_search(t_function func, t_bool repeat);
 static int enter_pager(int art, t_bool ignore_unavail);
 static int enter_thread(int depth, t_pagerinfo *page);
-static int find_new_pos(long old_artnum, int cur_pos);
+static int find_new_pos(t_artnum old_artnum, int cur_pos);
 static int group_catchup(t_function func);
 static int tab_pressed(void);
 static t_bool prompt_getart_limit(void);
@@ -87,10 +90,21 @@ show_tagged_lines(
 	void)
 {
 	int i, j;
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+	wchar_t *wtmp;
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 
 	for (i = grpmenu.first; i < grpmenu.first + NOTESLINES && i < grpmenu.max; ++i) {
-		if ((i != grpmenu.curr) && (j = line_is_tagged(base[i])))
+		if ((i != grpmenu.curr) && (j = line_is_tagged((int) base[i]))) {
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+			if ((wtmp = char2wchar_t(tin_ltoa(j, 3)))) {
+				mark_screen(i, mark_offset - (3 - art_mark_width), wtmp);
+				free(wtmp);
+			}
+#else
 			mark_screen(i, mark_offset - 2, tin_ltoa(j, 3));
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
+		}
 	}
 }
 
@@ -142,13 +156,16 @@ group_page(
 	char key[MAXKEYLEN];
 	int i, n, ii;
 	int thread_depth;	/* Starting depth in threads we enter */
-	t_artnum old_artnum = T_ARTNUM_CONST(0);
+	t_artnum old_artnum;
 	struct t_art_stat sbuf;
 	struct t_article *art;
 	t_bool flag;
 	t_bool xflag = FALSE;	/* 'X'-flag */
 	t_bool repeat_search;
 	t_function func;
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+	wchar_t *wtmp;
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 
 	/*
 	 * Set the group attributes
@@ -186,17 +203,17 @@ group_page(
 	clear_note_area();
 
 	if (group->attribute->auto_select) {
-		error_message(2, _(txt_autoselecting_articles), printascii(key, func_to_key(GROUP_MARK_UNSELECTED_ARTICLES_READ, group_keys)));
+		error_message(2, _(txt_autoselecting_articles), PrintFuncKey(key, GROUP_MARK_UNSELECTED_ARTICLES_READ, group_keys));
 		do_auto_select_arts();						/* 'X' command */
 		xflag = TRUE;
 	}
 
 	show_group_page();
 
-#	ifdef DEBUG
+#ifdef DEBUG
 	if (debug & DEBUG_NEWSRC)
 		debug_print_bitmap(group, NULL);
-#	endif /* DEBUG */
+#endif /* DEBUG */
 
 	/* reset ret_code */
 	ret_code = 0;
@@ -225,11 +242,11 @@ group_page(
 					prompt_item_num(func_to_key(func, group_keys), group->attribute->thread_articles == THREAD_NONE ? _(txt_select_art) : _(txt_select_thread));
 				break;
 
-#	ifndef NO_SHELL_ESCAPE
+#ifndef NO_SHELL_ESCAPE
 			case GLOBAL_SHELL_ESCAPE:
 				do_shell_escape();
 				break;
-#	endif /* !NO_SHELL_ESCAPE */
+#endif /* !NO_SHELL_ESCAPE */
 
 			case GLOBAL_FIRST_PAGE:		/* show first page of threads */
 				top_of_list();
@@ -320,12 +337,10 @@ group_page(
 
 			case GROUP_CANCEL:	/* cancel current basenote */
 				if (grpmenu.curr >= 0) {
-					if (can_post || group->attribute->mailing_list != NULL) {
-						int ret;
-
+					if (can_post || (group->attribute->mailing_list && *group->attribute->mailing_list)) {
 						n = (int) base[grpmenu.curr];
-						ret = art_open(TRUE, &arts[n], group, &pgart, TRUE, _(txt_reading_article));
-						if (ret != ART_UNAVAILABLE && ret != ART_ABORT && cancel_article(group, &arts[n], n))
+						ii = art_open(TRUE, &arts[n], group, &pgart, TRUE, _(txt_reading_article));
+						if (ii != ART_UNAVAILABLE && ii != ART_ABORT && cancel_article(group, &arts[n], n))
 							show_group_page();
 						art_close(&pgart);
 					} else
@@ -442,6 +457,11 @@ group_page(
 				show_group_page();
 				break;
 
+			case GLOBAL_CONNECTION_INFO:
+				show_connection_page();
+				show_group_page();
+				break;
+
 			case GLOBAL_TOGGLE_HELP_DISPLAY:		/* toggle mini help menu */
 				toggle_mini_help(GROUP_LEVEL);
 				show_group_page();
@@ -449,18 +469,19 @@ group_page(
 
 			case GLOBAL_TOGGLE_INVERSE_VIDEO:		/* toggle inverse video */
 				toggle_inverse_video();
+				need_parse_fmt |= GROUP_LEVEL;
 				show_group_page();
 				show_inverse_video_status();
 				break;
 
-#	ifdef HAVE_COLOR
+#ifdef HAVE_COLOR
 			case GLOBAL_TOGGLE_COLOR:
 				if (toggle_color()) {
 					show_group_page();
 					show_color_status();
 				}
 				break;
-#	endif /* HAVE_COLOR */
+#endif /* HAVE_COLOR */
 
 			case GROUP_MARK_THREAD_READ:			/* mark current thread/range/tagged threads as read */
 			case MARK_THREAD_UNREAD:				/* or unread */
@@ -579,7 +600,7 @@ group_page(
 					 * tinrc.getart_limit
 					 */
 					if (cmdline.args & CMDLINE_GETART_LIMIT)
-						cmdline.args &= ~CMDLINE_GETART_LIMIT;
+						cmdline.args ^= CMDLINE_GETART_LIMIT;
 					ret_code = GRP_NEXTUNREAD;
 				}
 				break;
@@ -590,7 +611,9 @@ group_page(
 
 			case GROUP_TAG_PARTS: /* tag all in order */
 				if (0 <= grpmenu.curr) {
-					if (tag_multipart(grpmenu.curr) != 0) {
+					if (tag_multipart((int) base[grpmenu.curr]) != 0) {
+						int old_num = num_of_tagged_arts;
+
 						/*
 						 * on success, move the pointer to the next
 						 * untagged article just for ease of use's sake
@@ -598,14 +621,17 @@ group_page(
 						n = grpmenu.curr;
 						update_group_page();
 						do {
-							n++;
+							++n;
 							n %= grpmenu.max;
 							if (arts[base[n]].tagged == 0) {
 								move_to_item(n);
 								break;
 							}
 						} while (n != grpmenu.curr);
-						info_message(_(txt_info_all_parts_tagged));
+						if (old_num < num_of_tagged_arts)
+							info_message(_(txt_info_all_parts_tagged));
+						else
+							info_message(_(txt_info_all_parts_untagged));
 					}
 				}
 				break;
@@ -649,15 +675,31 @@ group_page(
 								arts[ii].tagged = ++num_of_tagged_arts;
 						}
 					}
-					if ((ii = line_is_tagged(n)))
+					if ((ii = line_is_tagged(n))) {
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+						if ((wtmp = char2wchar_t(tin_ltoa(ii, 3)))) {
+							mark_screen(grpmenu.curr, mark_offset - (3 - art_mark_width), wtmp);
+							free(wtmp);
+						}
+#else
 						mark_screen(grpmenu.curr, mark_offset - 2, tin_ltoa(ii, 3));
-					else {
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
+					} else {
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+						wchar_t mark[] = { L'\0', L'\0' };
+#else
 						char mark[] = { '\0', '\0' };
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 
 						stat_thread(grpmenu.curr, &sbuf);
 						mark[0] = sbuf.art_mark;
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+						mark_screen(grpmenu.curr, mark_offset - (3 - art_mark_width), (const wchar_t *) L"   "); /* clear space used by tag numbering */
+						mark_screen(grpmenu.curr, mark_offset + (art_mark_width - wcwidth(mark[0])), mark);
+#else
 						mark_screen(grpmenu.curr, mark_offset - 2, "  "); /* clear space used by tag numbering */
 						mark_screen(grpmenu.curr, mark_offset, mark);
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 					}
 					if (tagged)
 						show_tagged_lines();
@@ -668,14 +710,13 @@ group_page(
 						draw_subject_arrow();
 
 					info_message(tagged ? _(txt_prefix_untagged) : _(txt_prefix_tagged), txt_thread_singular);
-
 				}
 				break;
 
 			case GROUP_TOGGLE_THREADING:		/* Cycle through the threading types */
-				group->attribute->thread_articles = (group->attribute->thread_articles + 1) % (THREAD_MAX + 1);
+				group->attribute->thread_articles = CAST_BITS((group->attribute->thread_articles + 1) % (THREAD_MAX + 1), thread_articles);
 				if (grpmenu.curr >= 0) {
-					i = base[grpmenu.curr];								/* Save a copy of current thread */
+					i = (int) base[grpmenu.curr];								/* Save a copy of current thread */
 					make_threads(group, TRUE);
 					find_base(group);
 					if ((grpmenu.curr = which_thread(i)) < 0)			/* Restore current position in group */
@@ -709,8 +750,11 @@ group_page(
 				break;
 
 			case GLOBAL_DISPLAY_POST_HISTORY:	/* display messages posted by user */
-				if (user_posted_messages())
+				if (post_hist_page()) {
+					if (grpmenu.curr == -1 && grpmenu.max > 0)
+						grpmenu.curr = 0;
 					show_group_page();
+				}
 				break;
 
 			case MARK_ARTICLE_UNREAD:		/* mark base article of thread unread */
@@ -772,16 +816,24 @@ group_page(
 				}
 				n = 0;
 				for_each_art_in_thread(i, grpmenu.curr) {
-					arts[i].selected = flag;
+					arts[i].selected = CAST_BOOL(flag);
 					++n;
 				}
 				assert(n > 0);
 				{
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+					wchar_t mark[] = { L'\0', L'\0' };
+#else
 					char mark[] = { '\0', '\0' };
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 
 					stat_thread(grpmenu.curr, &sbuf);
 					mark[0] = sbuf.art_mark;
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+					mark_screen(grpmenu.curr, mark_offset + (art_mark_width - wcwidth(mark[0])), mark);
+#else
 					mark_screen(grpmenu.curr, mark_offset, mark);
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 				}
 
 				show_group_title(TRUE);
@@ -810,15 +862,18 @@ group_page(
 
 			case GROUP_SELECT_PATTERN:	/* select matching patterns */
 				if (grpmenu.curr >= 0) {
-					char pat[128];
+					char pat[LEN];
 					char *prompt;
-					struct regex_cache cache = { NULL, NULL };
+					char *spat;
+					struct regex_cache cache = REGEX_CACHE_INITIALIZER;
 
-					prompt = fmt_string(_(txt_select_pattern), tinrc.default_select_pattern);
-					if (!(prompt_string_default(prompt, tinrc.default_select_pattern, _(txt_info_no_previous_expression), HIST_SELECT_PATTERN))) {
+					prompt = fmt_string(_(txt_select_pattern), BlankIfNull(tinrc.default_select_pattern));
+					if (!(spat = prompt_string_default(prompt, tinrc.default_select_pattern, _(txt_info_no_previous_expression), HIST_SELECT_PATTERN))) {
+						FreeIfNeeded(spat);
 						free(prompt);
 						break;
 					}
+					tinrc.default_select_pattern = spat;
 					free(prompt);
 
 					if (STRCMPEQ(tinrc.default_select_pattern, "*")) {	/* all */
@@ -829,7 +884,7 @@ group_page(
 					} else
 						snprintf(pat, sizeof(pat), REGEX_FMT, tinrc.default_select_pattern);
 
-					if (tinrc.wildcard && !(compile_regex(pat, &cache, PCRE_CASELESS)))
+					if (tinrc.wildcard && !(compile_regex(pat, &cache, REGEX_CASELESS)))
 						break;
 
 					flag = FALSE;
@@ -846,10 +901,8 @@ group_page(
 						show_group_title(TRUE);
 						update_group_page();
 					}
-					if (tinrc.wildcard) {
-						FreeIfNeeded(cache.re);
-						FreeIfNeeded(cache.extra);
-					}
+					if (tinrc.wildcard)
+						regex_cache_destroy(&cache);
 				}
 				break;
 
@@ -890,7 +943,7 @@ group_page(
 				break;
 
 			default:
-				info_message(_(txt_bad_command), printascii(key, func_to_key(GLOBAL_HELP, group_keys)));
+				info_message(_(txt_bad_command), PrintFuncKey(key, GLOBAL_HELP, group_keys));
 				break;
 		} /* switch(ch) */
 	} /* ret_code >= 0 */
@@ -902,6 +955,9 @@ group_page(
 
 	art_close(&pgart);				/* Close any open art */
 
+	FreeAndNull(grp_fmt.str);
+	FreeAndNull(grp_fmt.date_str);
+
 	curr_group = NULL;
 
 	return ret_code;
@@ -917,10 +973,14 @@ show_group_page(
 	signal_context = cGroup;
 	currmenu = &grpmenu;
 
+	if (!grp_fmt.str || (need_parse_fmt & GROUP_LEVEL)) {
+		parse_format_string(curr_group->attribute->group_format ? BlankIfNull(*curr_group->attribute->group_format) : "", &grp_fmt);
+		need_parse_fmt &= ~GROUP_LEVEL;
+		mark_offset = 0;
+	}
+
 	ClearScreen();
 	set_first_screen_item();
-	parse_format_string(curr_group->attribute->group_format, &grp_fmt);
-	mark_offset = 0;
 	show_group_title(FALSE);
 
 	for (i = grpmenu.first; i < grpmenu.first + NOTESLINES && i < grpmenu.max; ++i)
@@ -942,17 +1002,34 @@ update_group_page(
 	void)
 {
 	int i, j;
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+	wchar_t mark[] = { L'\0', L'\0' };
+	wchar_t *wtmp;
+#else
 	char mark[] = { '\0', '\0' };
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 	struct t_art_stat sbuf;
 
 	for (i = grpmenu.first; i < grpmenu.first + NOTESLINES && i < grpmenu.max; ++i) {
-		if ((j = line_is_tagged(base[i])))
+		if ((j = line_is_tagged((int) base[i]))) {
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+			if ((wtmp = char2wchar_t(tin_ltoa(j, 3)))) {
+				mark_screen(i, mark_offset - (3 - art_mark_width), wtmp);
+				free(wtmp);
+			}
+#else
 			mark_screen(i, mark_offset - 2, tin_ltoa(j, 3));
-		else {
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
+		} else {
 			stat_thread(i, &sbuf);
 			mark[0] = sbuf.art_mark;
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+			mark_screen(i, mark_offset - (3 - art_mark_width), (const wchar_t *) L"   ");	/* clear space used by tag numbering */
+			mark_screen(i, mark_offset + (art_mark_width - wcwidth(mark[0])), mark);
+#else
 			mark_screen(i, mark_offset - 2, "  ");	/* clear space used by tag numbering */
 			mark_screen(i, mark_offset, mark);
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 			if (sbuf.art_mark == tinrc.art_marked_selected)
 				draw_mark_selected(i);
 		}
@@ -972,10 +1049,13 @@ draw_subject_arrow(
 	draw_arrow_mark(INDEX_TOP + grpmenu.curr - grpmenu.first);
 
 	if (tinrc.info_in_last_line) {
+		int n;
 		struct t_art_stat statbuf;
 
 		stat_thread(grpmenu.curr, &statbuf);
-		info_message("%s", arts[(statbuf.unread ? next_unread(base[grpmenu.curr]) : base[grpmenu.curr])].subject);
+		if (!statbuf.unread || (n = next_unread((int) base[grpmenu.curr])) == -1)
+			n = (int) base[grpmenu.curr];
+		info_message("%s", arts[n].subject);
 	} else if (grpmenu.curr == grpmenu.max - 1)
 		info_message(_(txt_end_of_arts));
 }
@@ -1018,7 +1098,7 @@ toggle_read_unread(
 
 	if (grpmenu.curr >= 0) {
 		if (curr_group->attribute->show_only_unread_arts || new_responses(grpmenu.curr))
-			i = base[grpmenu.curr];
+			i = (int) base[grpmenu.curr];
 		else if ((n = prev_unread((int) base[grpmenu.curr])) >= 0)
 			i = n;
 		else if ((n = next_unread((int) base[grpmenu.curr])) >= 0)
@@ -1044,7 +1124,7 @@ toggle_read_unread(
  */
 static int
 find_new_pos(
-	long old_artnum,
+	t_artnum old_artnum,
 	int cur_pos)
 {
 	int i, pos;
@@ -1081,25 +1161,54 @@ void
 mark_screen(
 	int screen_row,
 	int screen_col,
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+	const wchar_t *value)
+#else
 	const char *value)
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 {
 	if (tinrc.draw_arrow) {
 		MoveCursor(INDEX2LNUM(screen_row), screen_col);
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+		my_fputws(value, stdout);
+#else
 		my_fputs(value, stdout);
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 		stow_cursor();
 		my_flush();
 	} else {
 #ifdef USE_CURSES
+#	if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+		char *tmp;
+#	endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 		int y, x;
+
 		getyx(stdscr, y, x);
+		if (x < 0)
+			x = 0;
+		if (y < 0)
+			y = 0;
+#	if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+		if ((tmp = wchar_t2char(value))) {
+			mvaddstr(INDEX2LNUM(screen_row), screen_col, tmp);
+			free(tmp);
+		}
+#	else
 		mvaddstr(INDEX2LNUM(screen_row), screen_col, value);
+#	endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 		MoveCursor(y, x);
 #else
 		int i;
+
 		for (i = 0; value[i] != '\0'; i++)
 			screen[INDEX2SNUM(screen_row)].col[screen_col + i] = value[i];
+
 		MoveCursor(INDEX2LNUM(screen_row), screen_col);
+#	if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+		my_fputws(value, stdout);
+#	else
 		my_fputs(value, stdout);
+#	endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 #endif /* USE_CURSES */
 		currmenu->draw_arrow();
 	}
@@ -1123,11 +1232,11 @@ build_multipart_header(
 	char *ss;
 
 	if (cmplen > maxlen)
-		strncpy(dest, src, maxlen);
+		strncpy(dest, src, (size_t) maxlen);
 	else {
-		strncpy(dest, src, cmplen);
+		strncpy(dest, src, (size_t) cmplen);
 		ss = dest + cmplen;
-		snprintf(ss, maxlen - cmplen, "(%s/%d)", mark, total);
+		snprintf(ss, (size_t) (maxlen - cmplen), "(%s/%d)", mark, total);
 	}
 }
 
@@ -1154,6 +1263,7 @@ build_sline(
 	int j, k, n;
 	size_t len;
 	struct t_art_stat sbuf;
+	t_bool tagged = FALSE;
 #if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
 	wchar_t *wtmp, *wtmp2;
 #else
@@ -1167,9 +1277,9 @@ build_sline(
 	 * make it the same size like in !USE_CURSES case to simplify the code
 	 */
 #	if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
-		buffer = my_malloc(cCOLS * MB_CUR_MAX + 2);
+		buffer = my_malloc((size_t) cCOLS * MB_CUR_MAX + 2);
 #	else
-		buffer = my_malloc(cCOLS + 2);
+		buffer = my_malloc((size_t) cCOLS + 2);
 #	endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 #else
 	buffer = screen[INDEX2SNUM(i)].col;
@@ -1210,10 +1320,10 @@ build_sline(
 #if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
 					if ((wtmp = char2wchar_t(buf)) != NULL) {
 						wtmp2 = wcspart(wtmp, grp_fmt.len_date_max, TRUE);
+						free(wtmp);
 						if (wcstombs(tmp, wtmp2, sizeof(tmp) - 1) != (size_t) -1)
 							strcat(buffer, tmp);
 
-						free(wtmp);
 						free(wtmp2);
 					}
 #else
@@ -1230,12 +1340,13 @@ build_sline(
 
 					if ((wtmp = char2wchar_t(tmp)) != NULL) {
 						wtmp2 = wcspart(wtmp, grp_fmt.len_from, TRUE);
-						if (wcstombs(tmp, wtmp2, sizeof(tmp) - 1) != (size_t) -1)
-							strcat(buffer, tmp);
-
 						free(wtmp);
-						free(wtmp2);
-					}
+					} else
+						wtmp2 = wcspart(L"", grp_fmt.len_from, TRUE);
+					if (wcstombs(tmp, wtmp2, sizeof(tmp) - 1) != (size_t) -1)
+						strcat(buffer, tmp);
+
+					free(wtmp2);
 #else
 					len_start = strwidth(buffer);
 					get_author(FALSE, &arts[j], buffer + strlen(buffer), grp_fmt.len_from);
@@ -1250,9 +1361,9 @@ build_sline(
 
 			case 'I':	/* initials */
 				len = MIN(grp_fmt.len_initials, sizeof(tmp) - 1);
-				get_initials(&arts[j], tmp, len);
+				get_initials(&arts[j], tmp, (int) len);
 				strcat(buffer, tmp);
-				if ((k = len - strwidth(tmp)) > 0) {
+				if ((k = (int) (len - (size_t) strwidth(tmp))) > 0) {
 					buf = buffer + strlen(buffer);
 					for (; k > 0; --k)
 						*buf++ = ' ';
@@ -1265,7 +1376,7 @@ build_sline(
 					strcat(buffer, tin_ltoa(arts[j].line_count, grp_fmt.len_linecnt));
 				else {
 					buf = buffer + strlen(buffer);
-					for (k = grp_fmt.len_linecnt; k > 1; --k)
+					for (k = (int) grp_fmt.len_linecnt; k > 1; --k)
 						*buf++ = ' ';
 					*buf++ = '?';
 					*buf = '\0';
@@ -1274,11 +1385,20 @@ build_sline(
 
 			case 'm':	/* article flags, tag number, or whatever */
 				if (!grp_fmt.mark_offset)
-					grp_fmt.mark_offset = mark_offset = strwidth(buffer) + 2;
-				if ((k = line_is_tagged(respnum)))
+					grp_fmt.mark_offset = (size_t) (mark_offset = strwidth(buffer) + 2);
+				if ((k = line_is_tagged(respnum))) {
 					STRCPY(tmp_buf, tin_ltoa(k, 3));
-				else
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+					if (art_mark_width > 1)
+						strcat(buffer, " ");
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
+					tagged = TRUE;
+				} else
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+					snprintf(tmp_buf, sizeof(tmp_buf), "%s%"T_CHAR_FMT, art_mark_width > wcwidth(sbuf.art_mark) ? "   " : "  ", (wint_t) sbuf.art_mark);
+#else
 					snprintf(tmp_buf, sizeof(tmp_buf), "  %c", sbuf.art_mark);
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 				strcat(buffer, tmp_buf);
 				break;
 
@@ -1287,7 +1407,7 @@ build_sline(
 				strncpy(tmp, arts[j].refptr ? arts[j].refptr->txt : "", len);
 				tmp[len] = '\0';
 				strcat(buffer, tmp);
-				if ((k = len - strwidth(tmp)) > 0) {
+				if ((k = (int) (len - (size_t) strwidth(tmp))) > 0) {
 					buf = buffer + strlen(buffer);
 					for (; k > 0; --k)
 						*buf++ = ' ';
@@ -1305,7 +1425,7 @@ build_sline(
 					strcat(buffer, tin_ltoa(n, grp_fmt.len_respcnt));
 				else {
 					buf = buffer + strlen(buffer);
-					for (k = grp_fmt.len_respcnt; k > 0; --k)
+					for (k = (int) grp_fmt.len_respcnt; k > 0; --k)
 						*buf++ = ' ';
 					*buf = '\0';
 				}
@@ -1319,19 +1439,21 @@ build_sline(
 				len = curr_group->attribute->show_author != SHOW_FROM_NONE ? grp_fmt.len_subj : grp_fmt.len_subj + grp_fmt.len_from;
 
 				if (sbuf.multipart_have > 1) /* We have a multipart msg so lets built our new header info. */
-					build_multipart_header(arts_sub, len, arts[j].subject, sbuf.multipart_compare_len, sbuf.multipart_have, sbuf.multipart_total);
+					build_multipart_header(arts_sub, (int) len, arts[j].subject, sbuf.multipart_compare_len, sbuf.multipart_have, sbuf.multipart_total);
 				else
 					STRCPY(arts_sub, arts[j].subject);
 
 #if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
 				if ((wtmp = char2wchar_t(arts_sub)) != NULL) {
 					wtmp2 = wcspart(wtmp, len, TRUE);
-					if (wcstombs(tmp, wtmp2, sizeof(tmp) - 1) != (size_t) -1)
-						strcat(buffer, tmp);
-
 					free(wtmp);
-					free(wtmp2);
-				}
+				} else
+					wtmp2 = wcspart(L"", len, TRUE);
+
+				if (wcstombs(tmp, wtmp2, sizeof(tmp) - 1) != (size_t) -1)
+					strcat(buffer, tmp);
+
+				free(wtmp2);
 #else
 				len_start = strwidth(buffer);
 				strncat(buffer, arts_sub, len);
@@ -1360,7 +1482,7 @@ build_sline(
 #ifdef USE_CURSES
 	free(buffer);
 #endif /* USE_CURSES */
-	if (sbuf.art_mark == tinrc.art_marked_selected)
+	if (!tagged && sbuf.art_mark == tinrc.art_marked_selected)
 		draw_mark_selected(i);
 }
 
@@ -1369,8 +1491,12 @@ static void
 show_group_title(
 	t_bool clear_title)
 {
-	char buf[LEN], tmp[LEN], flag;
-	int i, art_cnt = 0, recent_art_cnt = 0, selected_art_cnt = 0, read_selected_art_cnt = 0, killed_art_cnt = 0;
+	char buf[LEN], tmp[LEN], keyhelp[MAXKEYLEN], flag;
+	char *grpname;
+	int i, len, keyhelplen, art_cnt = 0, recent_art_cnt = 0, selected_art_cnt = 0, read_selected_art_cnt = 0, killed_art_cnt = 0;
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+	wchar_t *wtmp, *wtmp2;
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 
 	for_each_art(i) {
 		if (arts[i].thread == ART_EXPIRED)
@@ -1378,71 +1504,71 @@ show_group_title(
 
 		if (curr_group->attribute->show_only_unread_arts) {
 			if (arts[i].status != ART_READ) {
-				art_cnt++;
+				++art_cnt;
 				if (tinrc.recent_time && ((time((time_t *) 0) - arts[i].date) < (tinrc.recent_time * DAY)))
-					recent_art_cnt++;
+					++recent_art_cnt;
 			}
 			if (arts[i].killed == ART_KILLED_UNREAD)
-				killed_art_cnt++;
+				++killed_art_cnt;
 		} else {
-			art_cnt++;
+			++art_cnt;
 			if (tinrc.recent_time && ((time((time_t *) 0) - arts[i].date) < (tinrc.recent_time * DAY)))
-				recent_art_cnt++;
+				++recent_art_cnt;
 
 			if (arts[i].killed)
-				killed_art_cnt++;
+				++killed_art_cnt;
 		}
 		if (arts[i].selected) {
 			if (arts[i].status != ART_READ)
-				selected_art_cnt++;
+				++selected_art_cnt;
 			else
-				read_selected_art_cnt++;
+				++read_selected_art_cnt;
 		}
 	}
 
 	/*
 	 * build the group title
 	 */
-	/* group name and thread count */
-	snprintf(buf, sizeof(buf), "%s (%d%c",
-		curr_group->name, grpmenu.max,
-		*txt_threading[curr_group->attribute->thread_articles]);
+	/* thread count */
+	snprintf(buf, sizeof(buf), " (%d%c",
+		grpmenu.max, *txt_threading[curr_group->attribute->thread_articles]);
 
 	/* article count */
 	if ((cmdline.args & CMDLINE_GETART_LIMIT) ? cmdline.getart_limit : tinrc.getart_limit)
-		snprintf(tmp, sizeof(tmp), " %d/%d%c",
+		snprintf(tmp, sizeof(tmp), " %d/%d%"T_CHAR_FMT,
 			(cmdline.args & CMDLINE_GETART_LIMIT) ? cmdline.getart_limit : tinrc.getart_limit, art_cnt,
-			(curr_group->attribute->show_only_unread_arts ? tinrc.art_marked_unread : tinrc.art_marked_read));
+			(wint_t) (curr_group->attribute->show_only_unread_arts ? tinrc.art_marked_unread : tinrc.art_marked_read));
 	else
-		snprintf(tmp, sizeof(tmp), " %d%c",
+		snprintf(tmp, sizeof(tmp), " %d%"T_CHAR_FMT,
 			art_cnt,
-			(curr_group->attribute->show_only_unread_arts ? tinrc.art_marked_unread : tinrc.art_marked_read));
+			(wint_t) (curr_group->attribute->show_only_unread_arts ? tinrc.art_marked_unread : tinrc.art_marked_read));
 	if (sizeof(buf) > strlen(buf) + strlen(tmp))
 		strcat(buf, tmp);
 
 	/* selected articles */
 	if (curr_group->attribute->show_only_unread_arts)
-		snprintf(tmp, sizeof(tmp), " %d%c",
-			selected_art_cnt, tinrc.art_marked_selected);
+		snprintf(tmp, sizeof(tmp), " %d%"T_CHAR_FMT,
+			selected_art_cnt, (wint_t) tinrc.art_marked_selected);
 	else
-		snprintf(tmp, sizeof(tmp), " %d%c %d%c",
-			selected_art_cnt, tinrc.art_marked_selected,
-			read_selected_art_cnt, tinrc.art_marked_read_selected);
+		snprintf(tmp, sizeof(tmp), " %d%"T_CHAR_FMT" %d%"T_CHAR_FMT,
+			selected_art_cnt, (wint_t) tinrc.art_marked_selected,
+			read_selected_art_cnt, (wint_t) tinrc.art_marked_read_selected);
 	if (sizeof(buf) > strlen(buf) + strlen(tmp))
 		strcat(buf, tmp);
 
 	/* recent articles */
 	if (tinrc.recent_time) {
-		snprintf(tmp, sizeof(tmp), " %d%c",
-			recent_art_cnt, tinrc.art_marked_recent);
+		snprintf(tmp, sizeof(tmp), " %d%"T_CHAR_FMT,
+			recent_art_cnt, (wint_t) tinrc.art_marked_recent);
 
 		if (sizeof(buf) > strlen(buf) + strlen(tmp))
 			strcat(buf, tmp);
 	}
 
 	/* killed articles */
-	snprintf(tmp, sizeof(tmp), " %d%c",
-		killed_art_cnt, tinrc.art_marked_killed);
+	snprintf(tmp, sizeof(tmp), " %d%"T_CHAR_FMT,
+		killed_art_cnt, (wint_t) tinrc.art_marked_killed);
+
 	if (sizeof(buf) > strlen(buf) + strlen(tmp))
 		strcat(buf, tmp);
 
@@ -1454,11 +1580,59 @@ show_group_title(
 	if (sizeof(buf) > strlen(buf) + strlen(tmp))
 		strcat(buf, tmp);
 
+	/*
+	 * determine max len for centered group name
+	 */
+	if (tinrc.show_help_mail_sign != SHOW_SIGN_NONE) {
+		if (tinrc.show_help_mail_sign == SHOW_SIGN_MAIL)
+			len = cCOLS - (2 * strwidth(_(txt_you_have_mail))) - 2;
+		else {
+			PrintFuncKey(keyhelp, GLOBAL_HELP, group_keys);
+			keyhelplen = strwidth(keyhelp);
+			if (tinrc.show_help_mail_sign == SHOW_SIGN_HELP)
+				len = cCOLS - (2 * (strwidth(_(txt_type_h_for_help)) - 2 + keyhelplen)) - 2;
+			else
+				len = cCOLS - (2 * MAX(strwidth(_(txt_type_h_for_help)) - 2 + keyhelplen, strwidth(_(txt_you_have_mail)))) - 2;
+		}
+	} else
+		len = cCOLS - strwidth(buf) - 2;
+
+	if (len < 0)
+		len = 0;
+	/* group name */
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+	if ((wtmp = char2wchar_t(curr_group->name)) != NULL) {
+		if (tinrc.abbreviate_groupname)
+			wtmp2 = abbr_wcsgroupname(wtmp, len);
+		else
+			wtmp2 = wstrunc(wtmp, len);
+		free(wtmp);
+		grpname = wchar_t2char(wtmp2);
+		free(wtmp2);
+		if (grpname) {
+			STRCPY(tmp, grpname);
+			strcat(tmp, buf);
+			free(grpname);
+		} else {
+			STRCPY(tmp, buf);
+		}
+	} else
+		STRCPY(tmp, buf);
+#else
+	if (tinrc.abbreviate_groupname)
+		grpname = abbr_groupname(curr_group->name, len);
+	else
+		grpname = strunc(curr_group->name, len);
+	STRCPY(tmp, grpname);
+	strcat(tmp, buf);
+	free(grpname);
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
+
 	if (clear_title) {
 		MoveCursor(0, 0);
 		CleartoEOLN();
 	}
-	show_title(buf);
+	show_title(tmp);
 }
 
 
@@ -1513,7 +1687,7 @@ enter_pager(
 	t_pagerinfo page;
 
 	page.art = art;
-	page.ignore_unavail = ignore_unavail;
+	page.ignore_unavail = CAST_BOOL(ignore_unavail);
 
 	return enter_thread(0, &page);
 }
@@ -1642,13 +1816,9 @@ group_catchup(
 
 				case 1:						/* We caught up - advance group */
 					return GRP_NEXT;
-					/* NOTREACHED */
-					break;
 
 				default:					/* Just leave the group */
 					return GRP_EXIT;
-					/* NOTREACHED */
-					break;
 			}
 			/* FALLTHROUGH */
 		default:							/* Should not be here */
@@ -1667,8 +1837,9 @@ prompt_getart_limit(
 
 	clear_message();
 	if ((p = tin_getline(_(txt_enter_getart_limit), 2, NULL, 0, FALSE, HIST_OTHER)) != NULL) {
-		tinrc.getart_limit = atoi(p);
-		ret = TRUE;
+		tinrc.getart_limit = s2i(p, INT_MIN, INT_MAX);
+		if (!errno)
+			ret = TRUE;
 	}
 	clear_message();
 	return ret;
