@@ -3,35 +3,38 @@
  *  Module    : thread.c
  *  Author    : I. Lea
  *  Created   : 1991-04-01
- *  Updated   : 2016-01-26
+ *  Updated   : 2024-12-21
  *  Notes     :
  *
- * Copyright (c) 1991-2016 Iain Lea <iain@bricbrac.de>
+ * Copyright (c) 1991-2025 Iain Lea <iain@bricbrac.de>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
+ *
+ * 1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions and the following disclaimer.
+ *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote
- *    products derived from this software without specific prior written
- *    permission.
  *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
- * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
- * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
- * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
- * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
- * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * 3. Neither the name of the copyright holder nor the names of its
+ *    contributors may be used to endorse or promote products derived from
+ *    this software without specific prior written permission.
+ *
+ * * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
+ * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
+ * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+ * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
+ * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+ * POSSIBILITY OF SUCH DAMAGE.
  */
 
 
@@ -53,9 +56,13 @@ t_bool show_subject;
 /*
  * Local prototypes
  */
-static char get_art_mark(struct t_article *art);
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+	static wchar_t get_art_mark(struct t_article *art);
+#else
+	static char get_art_mark(struct t_article *art);
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 static int enter_pager(int art, t_bool ignore_unavail, int level);
-static int thread_catchup(t_function func, struct t_group *group);
+static int thread_catchup(t_function func, const struct t_group *group);
 static int thread_tab_pressed(void);
 static t_bool find_unexpired(struct t_msgid *ptr);
 static t_bool has_sibling(struct t_msgid *ptr);
@@ -74,7 +81,7 @@ static void update_thread_page(void);
  * thdmenu.max		Essentially = # threaded arts in current thread
  * thdmenu.first	Response # at top of screen
  */
-static t_menu thdmenu = {0, 0, 0, show_thread_page, draw_thread_arrow, draw_thread_item };
+static t_menu thdmenu = { 0, 0, 0, show_thread_page, draw_thread_arrow, draw_thread_item };
 
 /* TODO: find a better solution */
 static int ret_code = 0;		/* Set to < 0 when it is time to leave this menu */
@@ -82,7 +89,11 @@ static int ret_code = 0;		/* Set to < 0 when it is time to leave this menu */
 /*
  * returns the mark which should be used for this article
  */
-static char
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+	static wchar_t
+#else
+	static char
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 get_art_mark(
 	struct t_article *art)
 {
@@ -118,15 +129,18 @@ build_tline(
 	int l,
 	struct t_article *art)
 {
-	char mark = '\0';
-	int gap, fill, i;
-	size_t len, len_start, len_end;
+	int w;
+	size_t gap, fill, i, len, len_start, len_end;
 	struct t_msgid *ptr;
 	char *buffer, *buf;
 	char *fmt = thrd_fmt.str;
 	char tmp[LEN];
 #if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+	char markbuf[sizeof(wchar_t) + 4];
 	wchar_t *wtmp, *wtmp2;
+	wchar_t mark[] = { L'\0', L'\0' };
+#else
+	char mark[] = { '\0', '\0' };
 #endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 
 #ifdef USE_CURSES
@@ -135,9 +149,9 @@ build_tline(
 	 * make it the same size like in !USE_CURSES case to simplify some code
 	 */
 #	if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
-		buffer = my_malloc(cCOLS * MB_CUR_MAX + 2);
+		buffer = my_malloc(cCOLS > 0 ? (size_t) cCOLS * MB_CUR_MAX + 2 : 0);
 #	else
-		buffer = my_malloc(cCOLS + 2);
+		buffer = my_malloc(cCOLS > 0 ? (size_t) cCOLS + 2: 0);
 #	endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 #else
 	buffer = screen[INDEX2SNUM(l)].col;
@@ -167,10 +181,10 @@ build_tline(
 #if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
 					if ((wtmp = char2wchar_t(buf)) != NULL) {
 						wtmp2 = wcspart(wtmp, thrd_fmt.len_date_max, TRUE);
+						free(wtmp);
 						if (wcstombs(tmp, wtmp2, sizeof(tmp) - 1) != (size_t) -1)
 							strcat(buffer, tmp);
 
-						free(wtmp);
 						free(wtmp2);
 					}
 #else
@@ -186,30 +200,34 @@ build_tline(
 
 				if ((wtmp = char2wchar_t(tmp)) != NULL) {
 					wtmp2 = wcspart(wtmp, thrd_fmt.len_from, TRUE);
-					if (wcstombs(tmp, wtmp2, sizeof(tmp) - 1) != (size_t) -1)
-						strcat(buffer, tmp);
-
 					free(wtmp);
-					free(wtmp2);
-				}
+				} else
+					wtmp2 = wcspart(L"", thrd_fmt.len_from, TRUE);
+				if (wcstombs(tmp, wtmp2, sizeof(tmp) - 1) != (size_t) -1)
+					strcat(buffer, tmp);
+
+				free(wtmp2);
 #else
-				len_start = strwidth(buffer);
-				get_author(TRUE, art, buffer + strlen(buffer), thrd_fmt.len_from);
-				fill = thrd_fmt.len_from - (strwidth(buffer) - len_start);
-				gap = strlen(buffer);
-				for (i = 0; i < fill; i++)
-					buffer[gap + i] = ' ';
-				buffer[gap + fill] = '\0';
+				if (curr_group->attribute->show_author != SHOW_FROM_NONE) {
+					len_start = strwidth(buffer);
+					get_author(TRUE, art, buffer + strlen(buffer), thrd_fmt.len_from);
+					fill = thrd_fmt.len_from - (strwidth(buffer) - len_start);
+					gap = strlen(buffer);
+					for (i = 0; i < fill; i++)
+						buffer[gap + i] = ' ';
+					buffer[gap + fill] = '\0';
+				}
 #endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 				break;
 
 			case 'I':	/* initials */
 				len = MIN(thrd_fmt.len_initials, sizeof(tmp) - 1);
-				get_initials(art, tmp, len);
+				get_initials(art, tmp, (int) len);
 				strcat(buffer, tmp);
-				if ((i = len - strwidth(tmp)) > 0) {
+				w = strwidth(tmp);
+				if (w > 0 && len > (size_t) w) {
 					buf = buffer + strlen(buffer);
-					for (; i > 0; --i)
+					for (i = len - (size_t) w; i > 0; --i)
 						*buf++ = ' ';
 					*buf = '\0';
 				}
@@ -229,14 +247,27 @@ build_tline(
 
 			case 'm':	/* article flags, tag number, or whatever */
 				if (!thrd_fmt.mark_offset)
-					thrd_fmt.mark_offset = mark_offset = strwidth(buffer) + 2;
+					thrd_fmt.mark_offset = (size_t) (mark_offset = strwidth(buffer) + 2);
 				if (art->tagged) {
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+					if (art_mark_width > 1)
+						strcat(buffer, " ");
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 					strcat(buffer, tin_ltoa(art->tagged, 3));
-					mark = '\0';
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+					mark[0] = L'\0';
+#else
+					mark[0] = '\0';
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 				} else {
+					mark[0] = get_art_mark(art);
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+					snprintf(markbuf, sizeof(markbuf), "%s%"T_CHAR_FMT, art_mark_width > wcwidth(mark[0]) ? "   " : "  ", (wint_t) mark[0]);
+					strcat(buffer, markbuf);
+#else
 					strcat(buffer, "   ");
-					mark = get_art_mark(art);
-					buffer[strlen(buffer) - 1] = mark;		/* insert mark */
+					buffer[strlen(buffer) - 1] = mark[0];		/* insert mark */
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 				}
 				break;
 
@@ -245,9 +276,10 @@ build_tline(
 				strncpy(tmp, art->refptr ? art->refptr->txt : "", len);
 				tmp[len] = '\0';
 				strcat(buffer, tmp);
-				if ((i = len - strwidth(tmp)) > 0) {
+				w = strwidth(tmp);
+				if (w > 0 && len > (size_t) w) {
 					buf = buffer + strlen(buffer);
-					for (; i > 0; --i)
+					for (i = len - (size_t) w; i > 0; --i)
 						*buf++ = ' ';
 					*buf = '\0';
 				}
@@ -263,7 +295,7 @@ build_tline(
 
 			case 'T':	/* thread/subject */
 				len = curr_group->attribute->show_author != SHOW_FROM_NONE ? thrd_fmt.len_subj : thrd_fmt.len_subj + thrd_fmt.len_from;
-				len_start = strwidth(buffer);
+				len_start = (size_t) strwidth(buffer);
 
 				switch (curr_group->attribute->thread_articles) {
 					case THREAD_REFS:
@@ -274,14 +306,14 @@ build_tline(
 						 */
 
 						if (art->refptr) {
-							make_prefix(art->refptr, buffer + strlen(buffer), len);
+							make_prefix(art->refptr, buffer + strlen(buffer), (int) len);
 
-							len_end = strwidth(buffer);
+							len_end = (size_t) strwidth(buffer);
 
 							/*
 							 * Copy in the subject up to where the author (if any) starts
 							 */
-							gap = len - (len_end - len_start);
+							gap = (len - (len_end - len_start));
 
 							/*
 							 * Mutt-like thread tree. by sjpark@sparcs.kaist.ac.kr
@@ -290,16 +322,15 @@ build_tline(
 							if (gap > 0) {
 								for (ptr = art->refptr->parent; ptr && IS_EXPIRED(ptr); ptr = ptr->parent)
 									;
-
 								if (!(ptr && arts[ptr->article].subject == art->subject))
 #if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
 								{
 									if ((wtmp = char2wchar_t(art->subject)) != NULL) {
 										wtmp2 = wcspart(wtmp, gap, TRUE);
+										free(wtmp);
 										if (wcstombs(tmp, wtmp2, sizeof(tmp) - 1) != (size_t) -1)
 											strcat(buffer, tmp);
 
-										free(wtmp);
 										free(wtmp2);
 									}
 								}
@@ -317,17 +348,17 @@ build_tline(
 					case THREAD_SUBJ:
 					case THREAD_MULTI:
 					case THREAD_PERC:
-						len_end = strwidth(buffer);
-						gap = len - (len_end - len_start);
+						len_end = (size_t) strwidth(buffer);
+						gap = (len - (len_end - len_start));
 						if (gap > 0) {
 #if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
 							{
 								if ((wtmp = char2wchar_t(art->subject)) != NULL) {
 									wtmp2 = wcspart(wtmp, gap, TRUE);
+									free(wtmp);
 									if (wcstombs(tmp, wtmp2, sizeof(tmp) - 1) != (size_t) -1)
 										strcat(buffer, tmp);
 
-									free(wtmp);
 									free(wtmp2);
 								}
 							}
@@ -345,10 +376,10 @@ build_tline(
 				}
 
 				/* pad out */
-				fill = len - (strwidth(buffer) - len_start);
+				fill = (len - ((size_t) strwidth(buffer) - len_start));
 				gap = strlen(buffer);
 				for (i = 0; i < fill; i++)
-				buffer[gap + i] = ' ';
+					buffer[gap + i] = ' ';
 				buffer[gap + fill] = '\0';
 				break;
 
@@ -359,15 +390,10 @@ build_tline(
 	/* protect display from non-displayable characters (e.g., form-feed) */
 	convert_to_printable(buffer, FALSE);
 
-	if (!tinrc.strip_blanks) {
-		/* Pad to end of line so that inverse bar looks 'good' */
-		fill = cCOLS - strwidth(buffer);
-		gap = strlen(buffer);
-		for (i = 0; i < fill; i++)
-			buffer[gap + i] = ' ';
-
-		buffer[gap + fill] = '\0';
-	}
+#ifndef USE_CURSES
+	if (tinrc.strip_blanks)
+		strcat(strip_line(buffer), cCRLF);
+#endif /* !USE_CURSES */
 
 	WriteLine(INDEX2LNUM(l), buffer);
 
@@ -375,8 +401,9 @@ build_tline(
 	free(buffer);
 #endif /* USE_CURSES */
 
-	if (mark == tinrc.art_marked_selected)
+	if (mark[0] == tinrc.art_marked_selected)
 		draw_mark_selected(l);
+	my_flush();
 }
 
 
@@ -385,7 +412,6 @@ draw_thread_item(
 	int item)
 {
 	build_tline(item, &arts[find_response(thread_basenote, item)]);
-	return;
 }
 
 
@@ -412,7 +438,7 @@ thread_right(
  * Show current thread.
  * If threaded on Subject: show
  *   <respnum> <name>
- * If threaded on References: or Archive-name: show
+ * If threaded on References:
  *   <respnum> <subject> <name>
  * Return values:
  *		GRP_RETSELECT	Return to selection screen
@@ -427,12 +453,17 @@ thread_page(
 	struct t_group *group,
 	int respnum,				/* base[] article of thread to view */
 	int thread_depth,			/* initial depth in thread */
-	t_pagerinfo *page)			/* !NULL if we must go direct to the pager */
+	const t_pagerinfo *page)			/* !NULL if we must go direct to the pager */
 {
 	char key[MAXKEYLEN];
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+	wchar_t mark[] = { L'\0', L'\0' };
+	wchar_t *wtmp;
+#else
 	char mark[] = { '\0', '\0' };
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 	int i, n;
-	t_artnum old_artnum = T_ARTNUM_CONST(0);
+	t_artnum old_artnum;
 	t_bool repeat_search;
 	t_function func;
 
@@ -458,7 +489,8 @@ thread_page(
 	else {
 		if (group->attribute->pos_first_unread) {
 			if (new_responses(thread_basenote)) {
-				for (n = 0, i = (int) base[thread_basenote]; i >= 0; i = arts[i].thread, n++) {
+				n = 0;
+				for (i = (int) base[thread_basenote]; i >= 0; i = arts[i].thread, n++) {
 					if (arts[i].status == ART_UNREAD || arts[i].status == ART_WILL_RETURN) {
 						if (arts[i].thread == ART_EXPIRED)
 							art_mark(group, &arts[i], ART_READ);
@@ -599,31 +631,58 @@ thread_page(
 				break;
 
 			case GLOBAL_EDIT_FILTER:
-				if (invoke_editor(filter_file, filter_file_offset, NULL)) {
-					old_artnum = arts[find_response(thread_basenote, thdmenu.curr)].artnum;
-					unfilter_articles(group);
-					(void) read_filter_file(filter_file);
-					filter_articles(group);
-					make_threads(group, FALSE);
-					if ((n = find_artnum(old_artnum)) == -1 || which_thread(n) == -1) { /* We have lost the thread */
-						ret_code = GRP_KILLED;
-						break;
+				if (no_write)
+					info_message(_(txt_info_no_write));
+				else {
+					if (invoke_editor(filter_file, filter_file_offset, NULL)) {
+						old_artnum = arts[find_response(thread_basenote, thdmenu.curr)].artnum;
+						unfilter_articles(group);
+						(void) read_filter_file(filter_file);
+						filter_articles(group);
+						make_threads(group, FALSE);
+						if ((n = find_artnum(old_artnum)) == -1 || which_thread(n) == -1) { /* We have lost the thread */
+							ret_code = GRP_KILLED;
+							break;
+						}
+						fixup_thread(n, TRUE);
 					}
-					fixup_thread(n, TRUE);
+					show_thread_page();
 				}
-				show_thread_page();
 				break;
 
 			case THREAD_READ_ARTICLE:	/* read current article within thread */
 				ret_code = enter_pager(find_response(thread_basenote, thdmenu.curr), FALSE, THREAD_LEVEL);
 				break;
 
+/*
+			case THREAD_FOLLOWUP_QUOTE_HEADERS:
+			may need
+					if (func == THREAD_FOLLOWUP_QUOTE_HEADERS)
+						resize_article(TRUE, &pgart);
+			but as '^W' is already taken by MARK_FEED_UNREAD
+			we leave that function out for now
+*/
+			case THREAD_FOLLOWUP_QUOTE:
+			case THREAD_FOLLOWUP:
+				if (can_post || (group->attribute->mailing_list && *group->attribute->mailing_list)) {
+					int ret;
+
+					n = find_response(thread_basenote, thdmenu.curr);
+					ret = art_open(TRUE, &arts[n], group, &pgart, TRUE, _(txt_reading_article));
+					if (ret != ART_UNAVAILABLE && ret != ART_ABORT && n >= 0) {
+						post_response(group->name, n, (func == THREAD_FOLLOWUP_QUOTE) ? TRUE : FALSE, FALSE, FALSE);
+						show_thread_page();
+					}
+					art_close(&pgart);
+				}
+				break;
+
 			case THREAD_READ_NEXT_ARTICLE_OR_THREAD:
 				ret_code = thread_tab_pressed();
 				break;
 
 			case THREAD_CANCEL:		/* cancel current article */
-				if (can_post || group->attribute->mailing_list != NULL) {
+				if (can_post || (group->attribute->mailing_list && *group->attribute->mailing_list)) {
 					int ret;
 
 					n = find_response(thread_basenote, thdmenu.curr);
@@ -715,6 +774,11 @@ thread_page(
 				show_thread_page();
 				break;
 
+			case GLOBAL_CONNECTION_INFO:
+				show_connection_page();
+				show_thread_page();
+				break;
+
 			case GLOBAL_LOOKUP_MESSAGEID:
 				if ((n = prompt_msgid()) != ART_UNAVAILABLE)
 					ret_code = enter_pager(n, FALSE, THREAD_LEVEL);
@@ -746,6 +810,7 @@ thread_page(
 
 			case GLOBAL_TOGGLE_INVERSE_VIDEO:	/* toggle inverse video */
 				toggle_inverse_video();
+				need_parse_fmt |= THREAD_LEVEL;
 				show_thread_page();
 				show_inverse_video_status();
 				break;
@@ -767,6 +832,24 @@ thread_page(
 				ret_code = GRP_QUIT;
 				break;
 
+			case THREAD_TAG_PARTS:			/* tag/untag article */
+				/* Find index of current article */
+				if ((n = find_response(thread_basenote, thdmenu.curr)) < 0)
+					break;
+				else {
+					int old_num = num_of_tagged_arts;
+
+					if (tag_multipart(n) != 0) {
+						update_thread_page();
+
+						if (old_num < num_of_tagged_arts)
+							info_message(_(txt_info_all_parts_tagged));
+						else
+							info_message(_(txt_info_all_parts_untagged));
+					}
+				}
+				break;
+
 			case THREAD_TAG:			/* tag/untag article */
 				/* Find index of current article */
 				if ((n = find_response(thread_basenote, thdmenu.curr)) < 0)
@@ -774,9 +857,16 @@ thread_page(
 				else {
 					t_bool tagged;
 
-					if ((tagged = tag_article(n)))
+					if ((tagged = tag_article(n))) {
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+						if ((wtmp = char2wchar_t(tin_ltoa((&arts[n])->tagged, 3)))) {
+							mark_screen(thdmenu.curr, mark_offset - (3 - art_mark_width), wtmp);
+							free(wtmp);
+						}
+#else
 						mark_screen(thdmenu.curr, mark_offset - 2, tin_ltoa((&arts[n])->tagged, 3));
-					else
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
+					} else
 						update_thread_page();						/* Must update whole page */
 
 					/* Automatically advance to next art if not at end of thread */
@@ -815,7 +905,11 @@ thread_page(
 				arts[n].selected = (!(func == THREAD_TOGGLE_ARTICLE_SELECTION && arts[n].selected));	/* TODO: optimise? */
 /*				update_thread_page(); */
 				mark[0] = get_art_mark(&arts[n]);
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+				mark_screen(thdmenu.curr, mark_offset + (art_mark_width - wcwidth(mark[0])), mark);
+#else
 				mark_screen(thdmenu.curr, mark_offset, mark);
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 				if (thdmenu.curr + 1 < thdmenu.max)
 					move_down();
 				else
@@ -843,8 +937,11 @@ thread_page(
 				break;
 
 			case GLOBAL_DISPLAY_POST_HISTORY:	/* display messages posted by user */
-				if (user_posted_messages())
-					show_thread_page();
+				if (post_hist_page()) {
+					FreeAndNull(thrd_fmt.str);
+					FreeAndNull(thrd_fmt.date_str);
+					return GRP_EXIT;
+				}
 				break;
 
 			case GLOBAL_TOGGLE_INFO_LAST_LINE:		/* display subject in last line */
@@ -853,13 +950,16 @@ thread_page(
 				break;
 
 			default:
-				info_message(_(txt_bad_command), printascii(key, func_to_key(GLOBAL_HELP, thread_keys)));
+				info_message(_(txt_bad_command), PrintFuncKey(key, GLOBAL_HELP, thread_keys));
 		}
 	} /* ret_code >= 0 */
 
 	set_xclick_off();
 	clear_note_area();
 
+	FreeAndNull(thrd_fmt.str);
+	FreeAndNull(thrd_fmt.date_str);
+
 	return ret_code;
 }
 
@@ -869,25 +969,52 @@ show_thread_page(
 	void)
 {
 	char *title;
-	int i, art;
+	char keyhelp[MAXKEYLEN];
+	int i, art, keyhelplen, len;
 
 	signal_context = cThread;
 	currmenu = &thdmenu;
-	show_subject = FALSE;
+
+	if (!thrd_fmt.str || (need_parse_fmt & THREAD_LEVEL)) {
+		show_subject = FALSE;
+		parse_format_string(curr_group->attribute->thread_format ? BlankIfNull(*curr_group->attribute->thread_format) : "", &thrd_fmt);
+		need_parse_fmt &= ~THREAD_LEVEL;
+		mark_offset = 0;
+	}
 
 	ClearScreen();
 	set_first_screen_item();
 
-	parse_format_string(curr_group->attribute->thread_format, &thrd_fmt);
-	mark_offset = 0;
-
-	if (!show_subject)
-		show_subject = arts[thread_respnum].archive != NULL;
-
 	if (show_subject)
 		title = fmt_string(_(txt_stp_list_thread), grpmenu.curr + 1, grpmenu.max);
-	else
-		title = fmt_string(_(txt_stp_thread), cCOLS - 23, arts[thread_respnum].subject);
+	else {
+		/*
+		 * determine max. length for centered title
+		 * txt_stp_thread: "Thread (%.*s)"
+		 *                          ^^^^
+		 * Subtracting 4 from strwidth results in a trimmed long title with
+		 * a space on the right and no space on the left edge of the screen
+		 * --> subtracting only 3 makes it symmetrical
+		 */
+		if (tinrc.show_help_mail_sign != SHOW_SIGN_NONE) {
+			if (tinrc.show_help_mail_sign == SHOW_SIGN_MAIL)
+				len = cCOLS - (2 * strwidth(_(txt_you_have_mail))) - (strwidth(_(txt_stp_thread)) - 3);
+			else {
+				PrintFuncKey(keyhelp, GLOBAL_HELP, thread_keys);
+				keyhelplen = strwidth(keyhelp);
+				if (tinrc.show_help_mail_sign == SHOW_SIGN_HELP)
+					len = cCOLS - (2 * (strwidth(_(txt_type_h_for_help)) - 2 + keyhelplen)) - (strwidth(_(txt_stp_thread)) - 3);
+				else
+					len = cCOLS - (2 * MAX(strwidth(_(txt_type_h_for_help)) - 2 + keyhelplen, strwidth(_(txt_you_have_mail)))) - (strwidth(_(txt_stp_thread)) - 3);
+			}
+		} else
+			len = cCOLS - (strwidth(_(txt_stp_thread)) - 3);
+
+		if (len < 0)
+			len = 0;
+		title = fmt_string(_(txt_stp_thread), len, arts[thread_respnum].subject);
+	}
+
 	show_title(title);
 	free(title);
 
@@ -907,19 +1034,36 @@ static void
 update_thread_page(
 	void)
 {
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+	wchar_t mark[] = { L'\0', L'\0' };
+	wchar_t *wtmp;
+#else
 	char mark[] = { '\0', '\0' };
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 	int i, the_index;
 
 	the_index = find_response(thread_basenote, thdmenu.first);
 	assert(thdmenu.first != 0 || the_index == thread_respnum);
 
 	for (i = thdmenu.first; i < thdmenu.first + NOTESLINES && i < thdmenu.max; ++i) {
-		if ((&arts[the_index])->tagged)
+		if ((&arts[the_index])->tagged) {
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+			if ((wtmp = char2wchar_t(tin_ltoa((&arts[the_index])->tagged, 3)))) {
+				mark_screen(i, mark_offset - (3 - art_mark_width), wtmp);
+				free(wtmp);
+			}
+#else
 			mark_screen(i, mark_offset - 2, tin_ltoa((&arts[the_index])->tagged, 3));
-		else {
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
+		} else {
 			mark[0] = get_art_mark(&arts[the_index]);
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+			mark_screen(i, mark_offset - (3 - art_mark_width), (const wchar_t *) L"   ");	/* clear space used by tag numbering */
+			mark_screen(i, mark_offset + (art_mark_width - wcwidth(mark[0])), mark);
+#else
 			mark_screen(i, mark_offset - 2, "  ");	/* clear space used by tag numbering */
 			mark_screen(i, mark_offset, mark);
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 			if (mark[0] == tinrc.art_marked_selected)
 				draw_mark_selected(i);
 		}
@@ -959,7 +1103,7 @@ fixup_thread(
 	if (basenote >= 0) {
 		thread_basenote = basenote;
 		thdmenu.max = num_of_responses(thread_basenote) + 1;
-		thread_respnum = base[thread_basenote];
+		thread_respnum = (int) base[thread_basenote];
 		grpmenu.curr = basenote;
 		if (redraw && basenote != old_thread_basenote)
 			show_thread_page();
@@ -982,7 +1126,7 @@ new_responses(
 
 	for_each_art_in_thread(i, thread) {
 		if (arts[i].status != ART_READ)
-			sum++;
+			++sum;
 	}
 
 	return sum;
@@ -1041,7 +1185,7 @@ which_response(
 		if (j == n)
 			break;
 		else
-			num++;
+			++num;
 	}
 
 	return num;
@@ -1057,16 +1201,20 @@ num_of_responses(
 	int n)
 {
 	int i;
-	int oldi = -3;
 	int sum = 0;
+#ifndef NDEBUG
+	int oldi = -3;
 
 	assert(n < grpmenu.max && n >= 0);
+#endif /* !NDEBUG */
 
 	for_each_art_in_thread(i, n) {
+#ifndef NDEBUG
 		assert(i != ART_EXPIRED);
 		assert(i != oldi);
 		oldi = i;
-		sum++;
+#endif /* !NDEBUG */
+		++sum;
 	}
 
 	return sum - 1;
@@ -1102,7 +1250,7 @@ get_score_of_thread(
 			} else { /* tinrc.thread_score >= THREAD_SCORE_SUM */
 				/* sum scores of unread arts and count num. arts */
 				score += arts[i].score;
-				j++;
+				++j;
 			}
 		}
 	}
@@ -1122,7 +1270,7 @@ stat_thread(
 	struct t_art_stat *sbuf) /* return value is always ignored */
 {
 	int i;
-	MultiPartInfo minfo;
+	MultiPartInfo minfo = { NULL, 0, 0, 0, 0 };
 
 	sbuf->total = 0;
 	sbuf->unread = 0;
@@ -1230,9 +1378,8 @@ int
 next_thread(
 	int n)
 {
-	int i;
+	int i = which_thread(n) + 1;
 
-	i = which_thread(n) + 1;
 	if (i >= grpmenu.max)
 		return -1;
 
@@ -1254,10 +1401,16 @@ prev_response(
 	if (arts[n].prev >= 0)
 		return arts[n].prev;
 
-	i = which_thread(n) - 1;
-
-	if (i < 0)
+#if 0
+	if ((i = which_thread(n) - 1) < 0)
 		return -1;
+#else /* __COVERITY__ CID 1549087 */
+	i = which_thread(n);
+	if (i <= 0)
+		return -1;
+
+	--i;
+#endif /* 0 */
 
 	return find_response(i, num_of_responses(i));
 }
@@ -1271,9 +1424,7 @@ find_response(
 	int i,
 	int n)
 {
-	int j;
-
-	j = (int) base[i];
+	int j = (int) base[i];
 
 	while (n-- > 0 && arts[j].thread >= 0)
 		j = arts[j].thread;
@@ -1301,7 +1452,7 @@ next_unread(
 	}
 
 	if (curr_group->attribute->wrap_on_next_unread) {
-		n = base[0];
+		n = (int) base[0];
 		while (n != cur_base_art && n >= 0) {
 			if (((arts[n].status == ART_UNREAD) || (arts[n].status == ART_WILL_RETURN)) && arts[n].thread != ART_EXPIRED)
 				return n;
@@ -1380,7 +1531,8 @@ make_prefix(
 	for (ptr = art->parent; ptr; ptr = ptr->parent)
 		depth += (!IS_EXPIRED(ptr) ? 1 : 0);
 
-	if ((depth == 0) || (maxlen < 1)) {
+	/* the code below expects a maxlen of at least 3 */
+	if ((depth == 0) || (maxlen < 3)) {
 		prefix[0] = '\0';
 		return;
 	}
@@ -1394,14 +1546,14 @@ make_prefix(
 
 		while (prefix_ptr > maxlen - 2 - odd) {
 			if (depth_level < maxlen / 5)
-				depth_level++;
+				++depth_level;
 			prefix_ptr -= maxlen - depth_level - 2 - odd;
 			odd = (odd ? 0 : 1);
 		}
 	}
 
 #if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
-	buf = my_malloc(sizeof(wchar_t) * prefix_ptr + 3 * sizeof(wchar_t));
+	buf = my_malloc(sizeof(wchar_t) * (size_t) prefix_ptr + 3 * sizeof(wchar_t));
 	buf[prefix_ptr + 2] = (wchar_t) '\0';
 #else
 	buf = my_malloc(prefix_ptr + 3);
@@ -1411,7 +1563,7 @@ make_prefix(
 	buf[prefix_ptr] = TREE_HORIZ;
 	buf[--prefix_ptr] = (has_sibling(art) ? TREE_VERT_RIGHT : TREE_UP_RIGHT);
 
-	for (ptr = art->parent; prefix_ptr > 1; ptr = ptr->parent) {
+	for (ptr = art->parent; ptr && prefix_ptr > 1; ptr = ptr->parent) {
 		if (IS_EXPIRED(ptr))
 			continue;
 		buf[--prefix_ptr] = TREE_BLANK;
@@ -1433,7 +1585,6 @@ make_prefix(
 	prefix[maxlen] = '\0'; /* just in case strlen(buf) > maxlen */
 	free(buf);
 #endif /* MULTIBYTE_ABLE && !NO_LOCALE */
-	return;
 }
 
 
@@ -1447,7 +1598,7 @@ make_prefix(
 static int
 thread_catchup(
 	t_function func,
-	struct t_group *group)
+	const struct t_group *group)
 {
 	char buf[LEN];
 	int i, n;
@@ -1491,7 +1642,8 @@ thread_catchup(
 				default:				/* Just leave the group */
 					return GRP_EXIT;
 			}
-			/* FALLTHROUGH */
+			break;
+
 		default:
 			break;
 	}
@@ -1611,14 +1763,23 @@ thread_mark_postprocess(
 	t_function feed_type,
 	int respnum)
 {
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+	wchar_t mark[] = { L'\0', L'\0' };
+#else
 	char mark[] = { '\0', '\0' };
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 	int n;
 
 	switch (function) {
 		case (FEED_MARK_READ):
 			if (feed_type == FEED_ARTICLE) {
 				mark[0] = get_art_mark(&arts[respnum]);
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+				mark_screen(thdmenu.curr, mark_offset - (3 - art_mark_width), (const wchar_t *) L"   ");
+				mark_screen(thdmenu.curr, mark_offset + (art_mark_width - wcwidth(mark[0])), mark);
+#else
 				mark_screen(thdmenu.curr, mark_offset, mark);
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 			} else
 				show_thread_page();
 
@@ -1631,7 +1792,12 @@ thread_mark_postprocess(
 		case (FEED_MARK_UNREAD):
 			if (feed_type == FEED_ARTICLE) {
 				mark[0] = get_art_mark(&arts[respnum]);
+#if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
+				mark_screen(thdmenu.curr, mark_offset - (3 - art_mark_width), (const wchar_t *) L"   ");
+				mark_screen(thdmenu.curr, mark_offset + (art_mark_width - wcwidth(mark[0])), mark);
+#else
 				mark_screen(thdmenu.curr, mark_offset, mark);
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 				draw_thread_arrow();
 			} else
 				show_thread_page();
